/*
 * Copyright 2019-2021 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.springframework.aot.context.bootstrap.generator.bean.descriptor;

import org.junit.jupiter.api.Test;

import org.springframework.aop.framework.ProxyFactory;
import org.springframework.aot.context.bootstrap.generator.sample.SimpleConfiguration;
import org.springframework.aot.context.bootstrap.generator.sample.injection.FieldInjectionComponent;
import org.springframework.aot.context.bootstrap.generator.sample.injection.InjectionConfiguration;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionBuilder;
import org.springframework.beans.factory.support.DefaultListableBeanFactory;
import org.springframework.beans.factory.support.GenericBeanDefinition;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.core.env.Environment;
import org.springframework.util.ReflectionUtils;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatIllegalArgumentException;

/**
 * Tests for {@link DefaultBeanInstanceDescriptorFactory}.
 *
 * @author Stephane Nicoll
 */
class DefaultBeanInstanceDescriptorFactoryTests {

	@Test
	void createWithNullBeanDefinition() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		assertThatIllegalArgumentException().isThrownBy(() -> new DefaultBeanInstanceDescriptorFactory(beanFactory).create(null));
	}

	@Test
	void createWithNoMatchingInstanceCreator() {
		BeanDefinition beanDefinition = new RootBeanDefinition(String.class);
		beanDefinition.getConstructorArgumentValues().addIndexedArgumentValue(0, new Object());
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		assertThat(new DefaultBeanInstanceDescriptorFactory(beanFactory).create(beanDefinition)).isNull();
	}

	@Test
	void createWithNonRootBeanDefinition() {
		GenericBeanDefinition beanDefinition = new GenericBeanDefinition();
		beanDefinition.setBeanClass(SimpleConfiguration.class);
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", beanDefinition);
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertSimpleConfigurationDescriptor(descriptor);
	}

	@Test
	void createWithConstructor() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", BeanDefinitionBuilder.rootBeanDefinition(SimpleConfiguration.class).getBeanDefinition());
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertSimpleConfigurationDescriptor(descriptor);
	}

	private void assertSimpleConfigurationDescriptor(BeanInstanceDescriptor descriptor) {
		assertThat(descriptor).isNotNull();
		assertThat(descriptor.getUserBeanClass()).isEqualTo(SimpleConfiguration.class);
		assertThat(descriptor.getInstanceCreator()).isNotNull();
		assertThat(descriptor.getInstanceCreator().getMember()).isEqualTo(SimpleConfiguration.class.getDeclaredConstructors()[0]);
		assertThat(descriptor.getInjectionPoints()).isEmpty();
		assertThat(descriptor.getProperties()).isEmpty();
	}

	@Test
	void createWithMethodInjectionPoints() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", BeanDefinitionBuilder.rootBeanDefinition(InjectionConfiguration.class).getBeanDefinition());
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertInjectionConfiguration(descriptor);
	}

	@Test
	void createWithMethodInjectionPointsUsingCglibProxy() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", BeanDefinitionBuilder.rootBeanDefinition(
				createCglibProxyType(InjectionConfiguration.class)).getBeanDefinition());
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertInjectionConfiguration(descriptor);
	}

	private void assertInjectionConfiguration(BeanInstanceDescriptor descriptor) {
		assertThat(descriptor.getUserBeanClass()).isEqualTo(InjectionConfiguration.class);
		assertThat(descriptor.getInstanceCreator()).isNotNull();
		assertThat(descriptor.getInstanceCreator().getMember()).isEqualTo(InjectionConfiguration.class.getDeclaredConstructors()[0]);
		assertThat(descriptor.getInjectionPoints()).hasSize(2);
		assertThat(descriptor.getInjectionPoints()).anySatisfy((injectionPoint) -> {
			assertThat(injectionPoint.getMember()).isEqualTo(ReflectionUtils.findMethod(InjectionConfiguration.class, "setEnvironment", Environment.class));
			assertThat(injectionPoint.isRequired()).isTrue();
		});
		assertThat(descriptor.getInjectionPoints()).anySatisfy((injectionPoint) -> {
			assertThat(injectionPoint.getMember()).isEqualTo(ReflectionUtils.findMethod(InjectionConfiguration.class, "setBean", String.class));
			assertThat(injectionPoint.isRequired()).isFalse();
		});
		assertThat(descriptor.getProperties()).isEmpty();
	}

	@Test
	void createWithFieldInjectionPoints() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", BeanDefinitionBuilder.rootBeanDefinition(FieldInjectionComponent.class).getBeanDefinition());
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertFieldInjectionComponent(descriptor);
	}

	@Test
	void createWithFieldInjectionPointsUsingCglibProxy() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", BeanDefinitionBuilder.rootBeanDefinition(
				createCglibProxyType(FieldInjectionComponent.class)).getBeanDefinition());
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertFieldInjectionComponent(descriptor);
	}


	private void assertFieldInjectionComponent(BeanInstanceDescriptor descriptor) {
		assertThat(descriptor.getUserBeanClass()).isEqualTo(FieldInjectionComponent.class);
		assertThat(descriptor.getInstanceCreator()).isNotNull();
		assertThat(descriptor.getInstanceCreator().getMember()).isEqualTo(FieldInjectionComponent.class.getDeclaredConstructors()[0]);
		assertThat(descriptor.getInjectionPoints()).hasSize(2);
		assertThat(descriptor.getInjectionPoints()).anySatisfy((injectionPoint) -> {
			assertThat(injectionPoint.getMember()).isEqualTo(ReflectionUtils.findField(FieldInjectionComponent.class, "environment"));
			assertThat(injectionPoint.isRequired()).isTrue();
		});
		assertThat(descriptor.getInjectionPoints()).anySatisfy((injectionPoint) -> {
			assertThat(injectionPoint.getMember()).isEqualTo(ReflectionUtils.findField(FieldInjectionComponent.class, "bean"));
			assertThat(injectionPoint.isRequired()).isFalse();
		});
		assertThat(descriptor.getProperties()).isEmpty();
	}

	@Test
	void createWithProperties() {
		DefaultListableBeanFactory beanFactory = new DefaultListableBeanFactory();
		beanFactory.registerBeanDefinition("test", BeanDefinitionBuilder.rootBeanDefinition(InjectionConfiguration.class)
				.addPropertyValue("name", "Hello").addPropertyValue("counter", 42).getBeanDefinition());
		BeanInstanceDescriptor descriptor = createDescriptor(beanFactory, "test");
		assertThat(descriptor.getProperties()).hasSize(2);
		assertThat(descriptor.getProperties()).anySatisfy((propertyDescriptor) -> {
			assertThat(propertyDescriptor.getWriteMethod()).isEqualTo(ReflectionUtils.findMethod(InjectionConfiguration.class, "setName", String.class));
			assertThat(propertyDescriptor.getPropertyValue().getName()).isEqualTo("name");
		});
		assertThat(descriptor.getProperties()).anySatisfy((propertyDescriptor) -> {
			assertThat(propertyDescriptor.getWriteMethod()).isEqualTo(ReflectionUtils.findMethod(InjectionConfiguration.class, "setCounter", Integer.class));
			assertThat(propertyDescriptor.getPropertyValue().getName()).isEqualTo("counter");
		});
	}

	private BeanInstanceDescriptor createDescriptor(DefaultListableBeanFactory beanFactory, String beanName) {
		return new DefaultBeanInstanceDescriptorFactory(beanFactory).create(beanFactory.getMergedBeanDefinition(beanName));
	}

	private Class<?> createCglibProxyType(Class<?> target) {
		ProxyFactory proxyFactory = new ProxyFactory();
		proxyFactory.setTargetClass(target);
		return proxyFactory.getProxy().getClass();
	}

}
